﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;

namespace ThickeningDemo
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
            ofd = new OpenFileDialog();
            sfd = new SaveFileDialog();
            string filter = "Image|*.jpg;*.png;*.gif";
            ofd.Filter = filter;
            sfd.Filter = filter;
        }

        OpenFileDialog ofd;
        SaveFileDialog sfd;
        string image_path;
        Bitmap image, processed, display;

        private void thickenToolStripMenuItem_Click(object sender, EventArgs e)
        {
            processed = processed.Thickening();
            display = processed.ScaleImage(image_frame.Width, image_frame.Height);
            image_frame.Image = display;
        }

        private void saveToolStripMenuItem_Click(object sender, EventArgs e)
        {
            sfd.FileName = image_path;
            if (sfd.ShowDialog() == DialogResult.OK)
            {
                processed.Save(sfd.FileName);
            }
        }

        private void loadToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (ofd.ShowDialog() == DialogResult.OK)
            {
                image_path = ofd.FileName;
                image = (Bitmap)Bitmap.FromFile(image_path);
                processed = image;
                display = processed.ScaleImage(image_frame.Width, image_frame.Height);
                image_frame.Image = display;
            }
        }
    }

    public static class Ext
    {
        public static Bitmap ScaleImage(this Bitmap image, int width, int height)
        {
            float xratio = (float)width / image.Width;
            float yratio = (float)height / image.Height;
            float ratio = Math.Min(xratio, yratio);

            int w = (int)(image.Width * ratio);
            int h = (int)(image.Height * ratio);

            Bitmap result = new Bitmap(image, new Size(w, h));
            return result;
        }

        public static Bitmap Thickening(this Bitmap image)
        {
            int w = image.Width;
            int h = image.Height;

            BitmapData image_data = image.LockBits(
                new Rectangle(0, 0, w, h),
                ImageLockMode.ReadOnly,
                PixelFormat.Format8bppIndexed);

            int bytes = image_data.Stride * image_data.Height;
            byte[] buffer = new byte[bytes];
            byte[] result = new byte[bytes];

            Marshal.Copy(image_data.Scan0, buffer, 0, bytes);
            image.UnlockBits(image_data);

            int[][,] se =
            {
                Structuring_Element.left,
                Structuring_Element.topleft,
                Structuring_Element.top,
                Structuring_Element.topright,
                Structuring_Element.right,
                Structuring_Element.bottomright,
                Structuring_Element.bottom,
                Structuring_Element.bottomleft
            };

            for (int i = 0; i < bytes; i++)
            {
                buffer[i] = (byte)((buffer[i] == 0) ? 15 : 0);
            }

            for (int t = 0; t < 5; t++)
            {
                for (int i = 0; i < se.Length; i++)
                {
                    for (int x = 1; x < w - 1; x++)
                    {
                        for (int y = 1; y < h - 1; y++)
                        {
                            int position = x + y * image_data.Stride;
                            List<bool> change = new List<bool>();
                            for (int kx = -1; kx < 2; kx++)
                            {
                                for (int ky = -1; ky < 2; ky++)
                                {
                                    int se_pos = position + kx + ky * image_data.Stride;
                                    int se_opp = position + (kx * -1) + (ky * -1) * image_data.Stride;

                                    if (se_pos == position)
                                    {
                                        continue;
                                    }

                                    if (buffer[se_pos] > 0 && se[i][kx + 1, ky + 1] == 1 && buffer[se_opp] == 0 && se[i][(kx * -1) + 1, (ky * -1) + 1] == 0)
                                    {
                                        change.Add(true);
                                    }
                                }
                            }
                            if (change.Count == 3 && buffer[position] > 0)
                            {
                                buffer[position] = 0;
                            }
                        }
                    }
                }
            }

            for (int i = 0; i < bytes; i++)
            {
                buffer[i] = (byte)((buffer[i] == 0) ? 15 : 0);
            }

            Bitmap res_img = new Bitmap(w, h);
            BitmapData res_data = res_img.LockBits(
                new Rectangle(0, 0, w, h),
                ImageLockMode.WriteOnly,
                PixelFormat.Format8bppIndexed);
            Marshal.Copy(buffer, 0, res_data.Scan0, bytes);
            res_img.UnlockBits(res_data);

            return res_img;
        }
    }

    public static class Structuring_Element
    {
        public static int[,] left
        {
            get
            {
                return new int[,]
                {
                    { 1, 0, 0 },
                    { 1, 1, 0 },
                    { 1, 0, 0 }
                };
            }
        }

        public static int[,] topleft
        {
            get
            {
                return new int[,]
                {
                    { 1, 1, 0 },
                    { 1, 1, 0 },
                    { 0, 0, 0 }
                };
            }
        }

        public static int[,] top
        {
            get
            {
                return new int[,]
                {
                    { 1, 1, 1 },
                    { 0, 1, 0 },
                    { 0, 0, 0 }
                };
            }
        }
        public static int[,] topright
        {
            get
            {
                return new int[,]
                {
                    { 0, 1, 1 },
                    { 0, 1, 1 },
                    { 0, 0, 0 }
                };
            }
        }
        public static int[,] right
        {
            get
            {
                return new int[,]
                {
                    { 0, 0, 1 },
                    { 0, 1, 1 },
                    { 0, 0, 1 }
                };
            }
        }

        public static int[,] bottomright
        {
            get
            {
                return new int[,]
                {
                    { 0, 0, 0 },
                    { 0, 1, 1 },
                    { 0, 1, 1 }
                };
            }
        }
        public static int[,] bottom
        {
            get
            {
                return new int[,]
                {
                    { 0, 0, 0 },
                    { 0, 1, 0 },
                    { 1, 1, 1 }
                };
            }
        }

        public static int[,] bottomleft
        {
            get
            {
                return new int[,]
                {
                    { 0, 0, 0 },
                    { 1, 1, 0 },
                    { 1, 1, 0 }
                };
            }
        }
    }
}
